May 31, 1989

Don Plum, a colleague of mine, recently showed me a technique he
had developed for speeding up binary searches.  I was so fascinated
by it that I decided to describe it and make it available to your
readers.

Begin by setting up a structure to contain the entries to be searched.
This would contain the "key" to be looked for as well as two pointers
to the same type of structure.  One will eventually contain the
location of entries with lower values, the other, the higher values.
For example,

----------------------------------------------------------------------
#define BSRCHTBLMAX 1000

unsigned long numrecs;

struct BSRCHTBL
	{
	long keyvalue;
	/* any other pertinent data would go here */
	struct BSRCHTBL *lower;
	struct BSRCHTBL *higher;
	} bsrchtbl[BSRCHTBLMAX];
----------------------------------------------------------------------

Once this table has been filled with "numrecs" worth of "keyvalue"s and
sorted, the pointers are initialized as follows:

----------------------------------------------------------------------
int mid, x, y, z;
struct BSRCHTBL *midptr;

/*
The following initializes all of the even numbered table field
'lower' and 'higher' entries to contain pointers to the next lower
and next higher table elements in a binary search tree.  All of
the odd numbered entries' pointers contain NULL.  After the NULL
initialization loop, the outer portion of the next loop moves
through the table first by every second entry, then by every
fourth, then by every eighth, and so on.  The inner loop sets
each lower/higher pointer to one-half the outer step distance
(in numbers of structure entries) below/above the current position.
The check of 'z' prevents the higher pointer from ever being
beyond the end of the number of records actually in the table.
This results in a binary search routine which requires no subscript
computations and no limit checking.
*/

for (x = 0; x < numrecs; x++)
	bsrchtbl[x].lower = bsrchtbl[x].higher = NULL;
for (x = 1; x < numrecs; x <<= 1)
	for (y = (x << 1) - 1; y < numrecs; y += (x << 1))
		{
		mid = y;
		bsrchtbl[y].lower = bsrchtbl + y - x;
		z = x;
		while ((y + z) >= numrecs)
			z >>= 1;
		if (z)
			bsrchtbl[y].higher = bsrchtbl + y + z;
		}
midptr = bsrchtbl + mid;
----------------------------------------------------------------------

Now that the table is filled, it can be searched as follows:

----------------------------------------------------------------------
register struct BSRCHTBL *srchptr;
long key_to_find;

srchptr = midptr;
key_to_find = (value to search for);
while (srchptr != NULL)
	if (srchptr->keyvalue == key_to_find)
		break;
	else
		if (srchptr->keyvalue > key_to_find)
			srchptr = srchptr->lower;
		else
			srchptr = srchptr->higher;
if (srchptr != NULL)
	(the key_to_find was found);
else
	(the key_to_find was not found);
----------------------------------------------------------------------

And that's all there is to it.  We found a considerable performance
improvement over using the system supplied binary search routine.


Larry R. Waibel
427 Haverford Road
Ramona, Calif.
        92065
619/789-6722